@@ -8,23 +8,13 @@ use Protocol::DBus::Signature ();
 
 our $_ENDIAN_PACK;
 
-# Set this to get actual Perl filehandles in the
-# message body. XXX FIXME This is a very hacky way to do it!
-our $FILEHANDLES;
-
-# XXX FIXME Hackety-hack …
-our $PRESERVE_VARIANT_SIGNATURES;
-
 # for testing
 our $DICT_CANONICAL;
 
-our @_MARSHAL_FDS;
-
-# sig, data (array ref)
+# data (array ref, always), sig
 sub marshal_le {
     local $_ENDIAN_PACK = '<';
-    local @_MARSHAL_FDS;
-    return( _marshal(@_[0, 1]), \@_MARSHAL_FDS );
+    return _marshal(@_[0, 1]);
 }
 
 # buf, buf offset, sig
@@ -35,8 +25,7 @@ sub unmarshal_le {
 
 sub marshal_be {
     local $_ENDIAN_PACK = '>';
-    local @_MARSHAL_FDS;
-    return( _marshal(@_[0, 1]), \@_MARSHAL_FDS );
+    return _marshal(@_[0, 1]);
 }
 
 sub unmarshal_be {
@@ -58,10 +47,6 @@ sub _marshal {
 
         my $datum = $_data_are_not_list ? $data : $data->[$si];
 
-        if (!defined $datum) {
-            die "Undefined datum (SCT=“$sct”)!";
-        }
-
         # Arrays
         if (index($sct, 'a') == 0) {
             _marshal_array( $sct, $datum, $buf_sr);
@@ -84,27 +69,11 @@ sub _marshal {
 
         # Anything else is a basic type.
         else {
-            if ($sct eq 'h') {
-                my $fd = fileno($datum);
-                die "fileno($datum) returned undef!" if !defined $fd;
-
-                my ($idx) = grep { $_MARSHAL_FDS[$_] == $fd } 0 .. $#_MARSHAL_FDS;
-
-                if (!defined $idx) {
-                    $idx = @_MARSHAL_FDS;
-                    push @_MARSHAL_FDS, $fd;
-                }
-
-                $datum = $idx;
-            }
-            elsif ($sct eq 'o') {
+            if ($sct eq 'o') {
                 $datum =~ m<\A/(?:[A-Za-z0-9_]+(?:/[A-Za-z0-9_]+)*)?\z> or do {
                     die "Invalid object path: “$datum”";
                 };
             }
-            elsif ($sct eq 's') {
-                utf8::encode($datum);
-            }
 
             Protocol::DBus::Pack::align_str($$buf_sr, Protocol::DBus::Pack::ALIGNMENT()->{$sct});
 
@@ -270,20 +239,7 @@ sub _unmarshal_sct {
 
     my $val = unpack("\@$buf_offset ($pack_tmpl)$_ENDIAN_PACK", $$buf_sr);
 
-    my $strlen;
-
-    if ($FILEHANDLES && $sct_sig eq 'h') {
-        $val = $FILEHANDLES->[$val] || do {
-            warn "UNIX_FD ($val) received that doesn’t refer to a received file descriptor!\n";
-            $val;
-        };
-    }
-    elsif ($is_string) {
-        $strlen = length $val;
-        utf8::decode($val);
-    }
-
-    return ($val, $buf_offset - $buf_start + Protocol::DBus::Pack::WIDTH()->{$sct_sig} + ($is_string ? $strlen : 0));
+    return ($val, $buf_offset - $buf_start + Protocol::DBus::Pack::WIDTH()->{$sct_sig} + ($is_string ? length($val) : 0));
 }
 
 sub _unmarshal_variant {
@@ -299,10 +255,7 @@ sub _unmarshal_variant {
 
     (my $val, $len) = _unmarshal_sct( $buf_sr, $buf_offset, $sig );
 
-    return(
-        $PRESERVE_VARIANT_SIGNATURES ? bless( [ $sig => $val ], 'Protocol::DBus::Type::Variant' ) : $val,
-        $len + $buf_offset - $buf_start,
-    );
+    return( $val, $len + $buf_offset - $buf_start );
 }
 
 sub _get_pack_template {
@@ -375,6 +328,92 @@ sub _unmarshal_struct {
     return ($items_ar, ($buf_offset - $buf_start) + $len);
 }
 
+#----------------------------------------------------------------------
+# The logic below is unused. I was under the impression that I’d need a
+# way to determine if a message body’s length matches the given SIGNATURE,
+# but of course we don’t because the header includes the body length.
+#----------------------------------------------------------------------
+
+sub buffer_length_satisfies_signature_le {
+    local $_ENDIAN_PACK = '<';
+    return (_buffer_length_satisfies_signature(@_))[0];
+}
+
+sub buffer_length_satisfies_signature_be {
+    local $_ENDIAN_PACK = '>';
+    return (_buffer_length_satisfies_signature(@_))[0];
+}
+
+sub _buffer_length_satisfies_signature {
+    my ($buf, $buf_offset, $sig) = @_;
+
+    my $sig_offset = 0;
+
+    while ($buf_offset <= length($buf)) {
+
+        # We’re good if this passes because it means the buffer is longer
+        # than the passed-in signature needs it to be.
+        return (1, $buf_offset) if $sig_offset == length($sig);
+
+        my $sct_length = Protocol::DBus::Signature::get_sct_length($sig, $sig_offset);
+
+        my $next_sct = substr(
+            $sig,
+            $sig_offset,
+            $sct_length,
+        );
+
+        $sig_offset += $sct_length;
+
+        if ($next_sct eq 'v') {
+            my ($variant_sig, $len) = _unmarshal_sct($buf, $buf_offset, 'g');
+            $buf_offset += $len;
+
+            # This has to recurse and preserve the offset.
+            my ($ok, $new_offset) = _buffer_length_satisfies_signature( $buf, $buf_offset, $variant_sig );
+            return 0 if !$ok;
+            $buf_offset = $new_offset;
+        }
+
+        # signatures
+        elsif ($next_sct eq 'g') {
+            # 2 for the length byte and the trailing NUL
+            $buf_offset += 2 + unpack( "\@$buf_offset C", $buf )
+        }
+
+        # strings and object paths
+        elsif ( Protocol::DBus::Pack::STRING()->{$next_sct} ) {
+            _add_uint32_variant_length(\$buf, \$buf_offset);
+            $buf_offset++;  #trailing NUL
+        }
+
+        # numerics
+        elsif ( my $width = Protocol::DBus::Pack::WIDTH()->{$next_sct} ) {
+            $buf_offset += $width;
+        }
+
+        else {
+            my $char0 = substr($next_sct, 0, 1);
+
+            if ($char0 eq 'a') {
+                _add_uint32_variant_length(\$buf, \$buf_offset);
+            }
+            elsif ($char0 eq '(') {
+                Protocol::DBus::Pack::align( $buf_offset, 8 );
+
+                my ($ok, $new_offset) = _buffer_length_satisfies_signature( $buf, $buf_offset, substr($next_sct, 1, -1) );
+                return 0 if !$ok;
+                $buf_offset = $new_offset;
+            }
+            else {
+                die "unrecognized SCT: “$next_sct”";
+            }
+        }
+    }
+
+    return 0;
+}
+
 sub _add_uint32_variant_length {
     my ($buf_sr, $buf_offset_sr) = @_;
 
